from datetime import date

future_month_list = ["F", "G", "H", "J", "K", "M", "N", "Q", "U", "V", "X", "Z"]

short_month: tuple[str, ...] = (
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
)


def future_date_to_code(dte: date) -> str:
    """Obtain a future code from a date.

    For example december 2010 will result in Z10.
    """
    return "%s%s" % (future_month_list[dte.month - 1], str(dte.year)[2:])


future_month_dict = dict(
    (future_month_list[i], (i + 1, short_month[i])) for i in range(0, 12)
)



import click
import pandas as pd
from rich.console import Console

import ccy

from .console import df_to_rich


@click.group()
def ccys() -> None:
    """Currency commands."""


@ccys.command()
def show() -> None:
    """Show table with all currencies."""
    df = pd.DataFrame(ccy.dump_currency_table())
    console = Console()
    console.print(df_to_rich(df, exclude=("symbol_raw",)))



import time
from datetime import date, datetime
from typing import Any

try:
    from dateutil.parser import parse as date_from_string
except ImportError:  # noqa

    def date_from_string(dte):  # type: ignore
        raise NotImplementedError


def todate(val: Any) -> date:
    """Convert val to a datetime.date instance by trying several
    conversion algorithm.
    If it fails it raise a ValueError exception.
    """
    if not val:
        raise ValueError("Value not provided")
    if isinstance(val, datetime):
        return val.date()
    elif isinstance(val, date):
        return val
    else:
        try:
            ival = int(val)
            sval = str(ival)
            if len(sval) == 8:
                return yyyymmdd2date(val)
            elif len(sval) == 5:
                return juldate2date(val)
            else:
                raise ValueError
        except Exception:
            # Try to convert using the parsing algorithm
            try:
                return date_from_string(val).date()
            except Exception:
                raise ValueError("Could not convert %s to date" % val)


def date2timestamp(dte: date) -> float:
    return time.mktime(dte.timetuple())


def jstimestamp(dte: date) -> float:
    """Convert a date to a javascript timestamp.

    A Javascript timestamp is the number of milliseconds since
    January 1, 1970 00:00:00 UTC."""
    return 1000 * date2timestamp(dte)


def timestamp2date(tstamp: float) -> date:
    "Converts a unix timestamp to a Python datetime object"
    dt = datetime.fromtimestamp(tstamp)
    if not dt.hour + dt.minute + dt.second + dt.microsecond:
        return dt.date()
    else:
        return dt


def yyyymmdd2date(dte: float | int) -> date:
    try:
        y = int(dte // 10000)
        md = dte % 10000
        m = int(md // 100)
        d = int(md % 100)
        return date(y, m, d)
    except Exception:
        raise ValueError("Could not convert %s to date" % dte)


def date2yyyymmdd(dte: date) -> int:
    return dte.day + 100 * (dte.month + 100 * dte.year)


def juldate2date(val: float | int) -> date:
    """Convert from a Julian date/datetime to python date or datetime"""
    ival = int(val)
    dec = val - ival
    try:
        val4 = 4 * ival
        yd = val4 % 1461
        st = 1899
        if yd >= 4:
            st = 1900
        yd1 = yd - 241
        y = val4 // 1461 + st
        if yd1 >= 0:
            q = yd1 // 4 * 5 + 308
            qq = q // 153
            qr = q % 153
        else:
            q = yd // 4 * 5 + 1833
            qq = q // 153
            qr = q % 153
        m = qq % 12 + 1
        d = qr // 5 + 1
    except Exception:
        raise ValueError("Could not convert %s to date" % val)
    if dec:
        dec24 = 24 * dec
        hours = int(dec24)
        minutes = int(60 * (dec24 - hours))
        tot_seconds = 60 * (60 * (dec24 - hours) - minutes)
        seconds = int(tot_seconds)
        microseconds = int(1000000 * (tot_seconds - seconds))
        return datetime(y, m, d, hours, minutes, seconds, microseconds)
    else:
        return date(y, m, d)


def date2juldate(val: date) -> float:
    """Convert from a python date/datetime to a Julian date & time"""
    f = 12 * val.year + val.month - 22803
    fq = f // 12
    fr = f % 12
    dt = (fr * 153 + 302) // 5 + val.day + fq * 1461 // 4
    if isinstance(val, datetime):
        return (
            dt
            + (
                val.hour
                + (val.minute + (val.second + 0.000001 * val.microsecond) / 60.0) / 60.0
            )
            / 24.0
        )
    else:
        return dt



"""Day Counter for Counting time between 2 dates.
Implemented::

    * Actual 360
    * Actual 365
    * 30 / 360
    * Actual Actual
"""

from __future__ import annotations

from copy import copy
from datetime import date
from typing import Any
from ..dates.utils import date_diff

__all__ = ["getdc", "DayCounter", "alldc"]


def getdc(name: str) -> DayCounter:
    return _day_counters[name]()


def alldc() -> dict[str, DayCounterMeta]:
    global _day_counters
    return copy(_day_counters)


class DayCounterMeta(type):
    def __new__(cls, name: str, bases: Any, attrs: Any) -> DayCounterMeta:
        new_class = super(DayCounterMeta, cls).__new__(cls, name, bases, attrs)
        if name := getattr(new_class, "name", ""):
            _day_counters[name] = new_class
        return new_class


_day_counters: dict[str, DayCounterMeta] = {}


class DayCounter(metaclass=DayCounterMeta):
    name: str = ""

    def count(self, start: date, end: date) -> float:
        """Count the number of days between 2 dates"""
        return date_diff(end, start).total_seconds() / 86400

    def dcf(self, start: date, end: date) -> float:
        return self.count(start, end) / 360.0


class Act360(DayCounter):
    name = "ACT/360"


class Act365(DayCounter):
    name = "ACT/365"

    def dcf(self, start: date, end: date) -> float:
        return self.count(start, end) / 365.0


class Thirty360(DayCounter):
    name = "30/360"

    def dcf(self, start: date, end: date) -> float:
        d1 = min(start.day, 30)
        d2 = min(end.day, 30)
        return 360 * (end.year - start.year) + 30 * (end.month - start.month) + d2 - d1


class ActAct(DayCounter):
    name = "ACT/ACT"

    def dcf(self, start: date, end: date) -> float:
        return self.act_act_years(end) - self.act_act_years(start)

    def act_act_years(self, dt: date) -> float:
        y = dt.year
        days_in_year = 365 if y % 4 else 366
        dd = date_diff(dt, date(y, 1, 1)).total_seconds() / 86400
        return y + dd / days_in_year



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, timedelta
import holidays
import holidays.countries
import holidays.financial

isoweekend = frozenset((6, 7))
oneday = timedelta(days=1)

trading_centres: dict[str, TradingCentre] = {}


def prevbizday(dte: date, nd: int = 1, tcs: str | None = None) -> date:
    return centres(tcs).prevbizday(dte, nd)


def nextbizday(dte: date, nd: int = 1, tcs: str | None = None) -> date:
    return centres(tcs).nextbizday(dte, nd)


def centres(codes: str | None = None) -> TradingCentres:
    tcs = TradingCentres()
    if codes:
        lcs = codes.upper().replace(" ", "").split(",")
        for code in lcs:
            tc = trading_centres.get(code)
            if tc:
                tcs.centres[tc.code] = tc
    return tcs


@dataclass
class TradingCentre:
    code: str
    calendar: holidays.HolidayBase

    def isholiday(self, dte: date) -> bool:
        return dte in self.calendar


@dataclass
class TradingCentres:
    centres: dict[str, TradingCentre] = field(default_factory=dict)

    @property
    def code(self) -> str:
        return ",".join(sorted(self.centres))

    def isbizday(self, dte: date) -> bool:
        if dte.isoweekday() in isoweekend:
            return False
        for c in self.centres.values():
            if c.isholiday(dte):
                return False
        return True

    def nextbizday(self, dte: date, nd: int = 1) -> date:
        n = 0
        while not self.isbizday(dte):
            dte += oneday
        while n < nd:
            dte += oneday
            if self.isbizday(dte):
                n += 1
        return dte

    def prevbizday(self, dte: date, nd: int = 1) -> date:
        n = 0
        if nd < 0:
            return self.nextbizday(dte, -nd)
        else:
            while not self.isbizday(dte):
                dte -= oneday
            n = 0
            while n < nd:
                dte -= oneday
                if self.isbizday(dte):
                    n += 1
        return dte


trading_centres.update(
    (tc.code, tc)
    for tc in (
        TradingCentre(
            code="TGT",
            calendar=holidays.financial.european_central_bank.EuropeanCentralBank(),  # type: ignore [no-untyped-call]
        ),
        TradingCentre(
            code="LON",
            calendar=holidays.countries.united_kingdom.UnitedKingdom(),  # type: ignore [no-untyped-call]
        ),
        TradingCentre(
            code="NY",
            calendar=holidays.countries.united_states.UnitedStates(),  # type: ignore [no-untyped-call]
        ),
    )
)



"""Python currencies"""

__version__ = "1.7.1"


from .core.country import (
    CountryError,
    countries,
    country,
    country_map,
    countryccy,
    eurozone,
    print_eurozone,
    set_new_country,
)
from .core.currency import (
    ccypair,
    currency,
    currency_pair,
    currencydb,
    dump_currency_table,
)
from .core.daycounter import alldc, getdc
from .dates.converters import (
    date2juldate,
    date2timestamp,
    date2yyyymmdd,
    date_from_string,
    jstimestamp,
    juldate2date,
    timestamp2date,
    todate,
    yyyymmdd2date,
)
from .dates.futures import future_date_to_code, future_month_dict
from .dates.period import Period, period

__all__ = [
    "currency",
    "currencydb",
    "ccypair",
    "currency_pair",
    "dump_currency_table",
    #
    "getdc",
    "alldc",
    #
    "country",
    "countryccy",
    "set_new_country",
    "countries",
    "country_map",
    "CountryError",
    "eurozone",
    "print_eurozone",
    "future_date_to_code",
    "future_month_dict",
    "period",
    "Period",
    "todate",
    "date2timestamp",
    "timestamp2date",
    "yyyymmdd2date",
    "date2yyyymmdd",
    "juldate2date",
    "date2juldate",
    "date_from_string",
    "jstimestamp",
]


# Shortcuts
def cross(code: str) -> str:
    return currency(code).as_cross()


def crossover(code: str) -> str:
    return currency(code).as_cross("/")


def all() -> tuple[str, ...]:
    return tuple(currencydb())


def g7() -> tuple[str, ...]:
    return ("EUR", "GBP", "USD", "CAD")


def g10() -> tuple[str, ...]:
    return g7() + ("CHF", "SEK", "JPY")


def g10m() -> tuple[str, ...]:
    """modified g10 = G10 + AUD, NZD, NOK"""
    return g10() + ("AUD", "NZD", "NOK")



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from .currency import currencydb

if TYPE_CHECKING:
    pass

# Eurozone countries (officially the euro area)
# see http://en.wikipedia.org/wiki/Eurozone
# using ISO 3166-1 alpha-2 country codes
# see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
#
eurozone = tuple(
    ("AT BE CY DE EE ES FI FR GR HR IE IT LU LV LT MT NL PT SI SK").split(" ")
)


def print_eurozone() -> None:
    for c in sorted(country(c).name for c in eurozone):
        print(c)


_countries: dict[str, Country] = {}
_country_ccys: dict[str, str] = {}
_country_maps: dict[str, str] = {}


class Country(Protocol):
    alpha_2: str
    name: str
    alpha_3: str = ""


class CountryError(Exception):
    pass


def country(code: str) -> Country:
    cdb = countries()
    code = country_map(code)
    return cdb[code]


def countryccy(code: str) -> str:
    cdb = countryccys()
    code = str(code).upper()
    return cdb.get(code, "")


def countries() -> dict[str, Country]:
    """
    get country dictionary from pytz and add some extra.
    """
    global _countries
    if not _countries:
        try:
            import pycountry

            _countries = {country.alpha_2: country for country in pycountry.countries}
        except Exception:
            pass
    return _countries


def countryccys() -> dict[str, str]:
    """
    Create a dictionary with keys given by countries ISO codes and values
    given by their currencies
    """
    global _country_ccys
    if not _country_ccys:
        v: dict[str, str] = {}
        _country_ccys = v
        ccys = currencydb()
        for euc in eurozone:
            v[euc] = "EUR"
        for c in ccys.values():
            if c.default_country:
                v[c.default_country] = c.code
    return _country_ccys


def set_new_country(code: str, ccy: str, name: str) -> None:
    """
    Add new country code to database
    """
    code = str(code).upper()
    cdb = countries()
    if code in cdb:
        raise CountryError("Country %s already in database" % code)
    ccys = currencydb()
    ccy = str(ccy).upper()
    if ccy not in ccys:
        raise CountryError("Currency %s not in database" % ccy)
    # hacky - but best way I could find
    cdb[code] = type(cdb["IT"])(
        alpha_2=code,
        name=name,
        official_name=name,
    )  # type: ignore
    cccys = countryccys()
    cccys[code] = ccy


def country_map(code: str) -> str:
    """
    Country mapping
    """
    code = str(code).upper()
    global _country_maps
    return _country_maps.get(code, code)


# Add eurozone to list of Countries
set_new_country("EZ", "EUR", "Eurozone")
# lagacy - to remove
set_new_country("EU", "EUR", "Eurozone")



from datetime import date, datetime, timezone, timedelta


def utcnow() -> datetime:
    return datetime.now(timezone.utc)


def as_utc(dt: date | None = None) -> datetime:
    if dt is None:
        return utcnow()
    elif isinstance(dt, datetime):
        return dt.astimezone(timezone.utc)
    else:
        return datetime(dt.year, dt.month, dt.day, tzinfo=timezone.utc)


def date_diff(a: date, b: date) -> timedelta:
    if isinstance(a, datetime) or isinstance(b, datetime):
        return as_utc(a) - as_utc(b)
    return a - b



import pandas as pd
from rich.table import Table

set_alike = set[str] | frozenset[str] | tuple[str] | list[str] | None


def df_to_rich(
    df: pd.DataFrame, *, exclude: set_alike = None, **columns: dict
) -> Table:
    table = Table()
    if exclude_columns := set(exclude or ()):
        df = df.drop(columns=exclude_columns)
    for column in df.columns:
        config = dict(justify="right", style="cyan", no_wrap=True)
        config.update(columns.get(column) or {})
        table.add_column(column, **config)  # type: ignore[arg-type]
    for row in df.values:
        table.add_row(*[str(item) for item in row])
    return table



from __future__ import annotations

from typing import Any


def period(pstr: str = "") -> Period:
    """Create a period object from a period string"""
    return Period.make(pstr)


def find_first_of(st: str, possible: str) -> int:
    lowi = -1
    for p in tuple(possible):
        i = st.find(p)
        if i != -1 and (i < lowi or lowi == -1):
            lowi = i
    return lowi


def safediv(x: int, d: int) -> int:
    return x // d if x >= 0 else -(-x // d)


def safemod(x: int, d: int) -> int:
    return x % d if x >= 0 else -(-x % d)


class Period:
    def __init__(self, months: int = 0, days: int = 0) -> None:
        self._months = months
        self._days = days

    @classmethod
    def make(cls, data: Any) -> Period:
        if isinstance(data, cls):
            return data
        elif isinstance(data, str):
            return cls().add_tenure(data)
        else:
            raise TypeError("Cannot convert %s to Period" % data)

    def isempty(self) -> bool:
        return self._months == 0 and self._days == 0

    def add_days(self, days: int) -> None:
        self._days += days

    def add_weeks(self, weeks: int) -> None:
        self._days += int(7 * weeks)

    def add_months(self, months: int) -> None:
        self._months += months

    def add_years(self, years: int) -> None:
        self._months += int(12 * years)

    @property
    def years(self) -> int:
        return safediv(self._months, 12)

    @property
    def months(self) -> int:
        return safemod(self._months, 12)

    @property
    def weeks(self) -> int:
        return safediv(self._days, 7)

    @property
    def days(self) -> int:
        return safemod(self._days, 7)

    @property
    def totaldays(self) -> int:
        return 30 * self._months + self._days

    def __repr__(self) -> str:
        """The period string"""
        return self.components()

    def __str__(self) -> str:
        return self.__repr__()

    def components(self) -> str:
        """The period string"""
        p = ""
        neg = self.totaldays < 0
        y = self.years
        m = self.months
        w = self.weeks
        d = self.days
        if y:
            p = "%sY" % abs(y)
        if m:
            p = "%s%sM" % (p, abs(m))
        if w:
            p = "%s%sW" % (p, abs(w))
        if d:
            p = "%s%sD" % (p, abs(d))
        return "-" + p if neg else p

    def simple(self) -> str:
        """A string representation with only one period delimiter."""
        if self._days:
            return "%sD" % self.totaldays
        elif self.months:
            return "%sM" % self._months
        elif self.years:
            return "%sY" % self.years
        else:
            return ""

    def add_tenure(self, pstr: str) -> Period:
        if isinstance(pstr, self.__class__):
            self._months += pstr._months
            self._days += pstr._days
            return self
        st = str(pstr).upper()
        done = False
        sign = 1
        while not done:
            if not st:
                done = True
            else:
                ip = find_first_of(st, "DWMY")
                if ip == -1:
                    raise ValueError("Unknown period %s" % pstr)
                p = st[ip]
                v = int(st[:ip])
                sign = sign if v > 0 else -sign
                v = sign * abs(v)
                if p == "D":
                    self.add_days(v)
                elif p == "W":
                    self.add_weeks(v)
                elif p == "M":
                    self.add_months(v)
                elif p == "Y":
                    self.add_years(v)
                ip += 1
                st = st[ip:]
        return self

    def __add__(self, other: Any) -> Period:
        p = self.make(other)
        return self.__class__(self._months + p._months, self._days + p._days)

    def __radd__(self, other: Any) -> Period:
        return self + other

    def __sub__(self, other: Any) -> Period:
        p = self.make(other)
        return self.__class__(self._months - p._months, self._days - p._days)

    def __rsub__(self, other: Any) -> Period:
        return self.make(other) - self

    def __gt__(self, other: Any) -> bool:
        return self.totaldays > self.make(other).totaldays

    def __lt__(self, other: Any) -> bool:
        return self.totaldays < self.make(other).totaldays

    def __ge__(self, other: Any) -> bool:
        return self.totaldays >= self.make(other).totaldays

    def __le__(self, other: Any) -> bool:
        return self.totaldays <= self.make(other).totaldays

    def __eq__(self, other: Any) -> bool:
        return self.totaldays == self.make(other).totaldays



from __future__ import annotations

import sys
from typing import Any, Callable, NamedTuple

from .data import make_ccys

usd_order = 5


def to_string(v: Any) -> str:
    if isinstance(v, bytes):
        return v.decode("utf-8")
    else:
        return "%s" % v


def overusdfun(v1: float) -> float:
    return v1


def overusdfuni(v1: float) -> float:
    return 1.0 / v1


class ccy(NamedTuple):
    code: str
    isonumber: str
    twoletterscode: str
    order: int
    name: str
    rounding: int
    default_country: str
    fixeddc: str = "Act/365"
    floatdc: str = "Act/365"
    fixedfreq: str = ""
    floatfreq: str = ""
    future: str = ""
    symbol_raw: str = r"\00a4"
    html: str = ""

    @property
    def symbol(self) -> str:
        return self.symbol_raw.encode("utf-8").decode("unicode_escape")

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ccy):
            return other.code == self.code
        return False

    def description(self) -> str:
        if self.order > usd_order:
            v = "USD / %s" % self.code
        else:
            v = "%s / USD" % self.code
        if self.order != usd_order:
            return "%s Spot Exchange Rate" % v
        else:
            return "Dollar"

    def info(self) -> dict[str, Any]:
        data = self._asdict()
        data["symbol"] = self.symbol
        return data

    def printinfo(self, stream: Any | None = None) -> None:
        info = self.info()
        stream = stream or sys.stdout
        for k, v in info.items():
            stream.write(to_string("%s: %s\n" % (k, v)))

    def __str__(self) -> str:
        return self.code

    def swap(self, c2: ccy) -> tuple[bool, ccy, ccy]:
        """
        put the order of currencies as market standard
        """
        inv = False
        c1 = self
        if c1.order > c2.order:
            ct = c1
            c1 = c2
            c2 = ct
            inv = True
        return inv, c1, c2

    def overusdfunc(self) -> Callable[[float], float]:
        if self.order > usd_order:
            return overusdfuni
        else:
            return overusdfun

    def usdoverfunc(self) -> Callable[[float], float]:
        if self.order > usd_order:
            return overusdfun
        else:
            return overusdfuni

    def as_cross(self, delimiter: str = "") -> str:
        """
        Return a cross rate representation with respect USD.
        @param delimiter: could be '' or '/' normally
        """
        if self.order > usd_order:
            return "USD%s%s" % (delimiter, self.code)
        else:
            return "%s%sUSD" % (self.code, delimiter)

    def spot(self, c2: ccy, v1: float, v2: float) -> float:
        if self.order > c2.order:
            vt = v1
            v1 = v2
            v2 = vt
        return v1 / v2


class ccy_pair(NamedTuple):
    """
    Currency pair such as EURUSD, USDCHF

    XXXYYY - XXX is the foreign currency, while YYY is the base currency

    XXXYYY means 1 unit of of XXX cost XXXYYY units of YYY
    """

    ccy1: ccy
    ccy2: ccy

    @property
    def code(self) -> str:
        return "%s%s" % (self.ccy1, self.ccy2)

    def __repr__(self) -> str:
        return "%s: %s" % (self.__class__.__name__, self.code)

    def __str__(self) -> str:
        return self.code

    def mkt(self) -> ccy_pair:
        if self.ccy1.order > self.ccy2.order:
            return ccy_pair(self.ccy2, self.ccy1)
        else:
            return self

    def over(self, name: str = "usd") -> ccy_pair:
        """Returns a new currency pair with the *over* currency as
        second part of the pair (Foreign currency)."""
        name = name.upper()
        if self.ccy1.code == name.upper():
            return ccy_pair(self.ccy2, self.ccy1)
        else:
            return self


class ccydb(dict[str, ccy]):
    def insert(self, *args: Any, **kwargs: Any) -> None:
        c = ccy(*args, **kwargs)
        self[c.code] = c


def currencydb() -> ccydb:
    global _ccys
    if not _ccys:
        _ccys = ccydb()
        make_ccys(_ccys)
    return _ccys


def ccypairsdb() -> dict[str, ccy_pair]:
    global _ccypairs
    if not _ccypairs:
        _ccypairs = make_ccypairs()
    return _ccypairs


def currency(code: str | ccy) -> ccy:
    c = currencydb()
    return c[str(code).upper()]


def ccypair(code: str | ccy_pair) -> ccy_pair:
    c = ccypairsdb()
    return c[str(code).upper()]


def currency_pair(code: str | ccy_pair) -> ccy_pair:
    """Construct a :class:`ccy_pair` from a six letter string."""
    c = str(code)
    c1 = currency(c[:3])
    c2 = currency(c[3:])
    return ccy_pair(c1, c2)


def make_ccypairs() -> dict[str, ccy_pair]:
    ccys = currencydb()
    db = {}

    for ccy1 in ccys.values():
        od = ccy1.order
        for ccy2 in ccys.values():
            if ccy2.order <= od:
                continue
            p = ccy_pair(ccy1, ccy2)
            db[p.code] = p
    return db


def dump_currency_table() -> list:
    return [c.info() for c in sorted(currencydb().values(), key=lambda x: x.order)]


_ccys: ccydb = ccydb()
_ccypairs: dict[str, ccy_pair] = {}



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .currency import ccydb


def make_ccys(db: ccydb) -> None:
    """
    Create the currency dictionary
    """
    dfr = 4
    dollar = r"\u0024"
    peso = r"\u20b1"
    kr = r"kr"
    insert = db.insert

    # G10 & SCANDI
    insert(
        "EUR",
        "978",
        "EU",
        1,
        "Euro",
        dfr,
        "EU",
        "30/360",
        "ACT/360",
        future="FE",
        symbol_raw=r"\u20ac",
        html="&#x20ac;",
    )
    insert(
        "GBP",
        "826",
        "BP",
        2,
        "British Pound",
        dfr,
        "GB",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u00a3",
        html="&#xa3;",
    )
    insert(
        "AUD",
        "036",
        "AD",
        3,
        "Australian Dollar",
        dfr,
        "AU",
        "ACT/365",
        "ACT/365",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "NZD",
        "554",
        "ND",
        4,
        "New-Zealand Dollar",
        dfr,
        "NZ",
        "ACT/365",
        "ACT/365",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "USD",
        "840",
        "UD",
        5,
        "US Dollar",
        0,
        "US",
        "30/360",
        "ACT/360",
        future="ED",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "CAD",
        "124",
        "CD",
        6,
        "Canadian Dollar",
        dfr,
        "CA",
        "ACT/365",
        "ACT/365",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "CHF",
        "756",
        "SF",
        7,
        "Swiss Franc",
        dfr,
        "CH",
        "30/360",
        "ACT/360",
        symbol_raw=r"Fr",
        html="&#x20a3;",
    )
    insert(
        "NOK",
        "578",
        "NK",
        8,
        "Norwegian Krona",
        dfr,
        "NO",
        "30/360",
        "ACT/360",
        symbol_raw=kr,
        html=kr,
    )
    insert(
        "SEK",
        "752",
        "SK",
        9,
        "Swedish Krona",
        dfr,
        "SE",
        "30/360",
        "ACT/360",
        symbol_raw=kr,
        html=kr,
    )
    insert(
        "DKK",
        "208",
        "DK",
        10,
        "Danish Krona",
        dfr,
        "DK",
        "30/360",
        "ACT/360",
        symbol_raw=kr,
        html=kr,
    )
    insert(
        "JPY",
        "392",
        "JY",
        10000,
        "Japanese Yen",
        2,
        "JP",
        "ACT/365",
        "ACT/360",
        symbol_raw=r"\u00a5",
        html="&#xa5;",
    )

    # ASIA
    insert(
        "CNY",
        "156",
        "CY",
        680,
        "Chinese Renminbi",
        dfr,
        "CN",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u00a5",
        html="&#xa5;",
    )
    insert(
        "KRW",
        "410",
        "KW",
        110000,
        "South Korean won",
        2,
        "KR",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u20a9",
        html="&#x20a9;",
    )
    insert(
        "SGD",
        "702",
        "SD",
        15,
        "Singapore Dollar",
        dfr,
        "SG",
        "ACT/365",
        "ACT/365",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "IDR",
        "360",
        "IH",
        970000,
        "Indonesian Rupiah",
        0,
        "ID",
        "ACT/360",
        "ACT/360",
        symbol_raw=r"Rp",
        html="Rp",
    )
    insert(
        "THB",
        "764",
        "TB",
        3300,
        "Thai Baht",
        2,
        "TH",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u0e3f",
        html="&#xe3f;",
    )
    insert(
        "TWD",
        "901",
        "TD",
        18,
        "Taiwan Dollar",
        dfr,
        "TW",
        "ACT/365",
        "ACT/365",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "HKD",
        "344",
        "HD",
        19,
        "Hong Kong Dollar",
        dfr,
        "HK",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u5713",
        html="HK&#x24;",
    )
    insert(
        "PHP",
        "608",
        "PP",
        4770,
        "Philippines Peso",
        dfr,
        "PH",
        "ACT/360",
        "ACT/360",
        symbol_raw=peso,
        html="&#x20b1;",
    )
    insert(
        "INR",
        "356",
        "IR",
        4500,
        "Indian Rupee",
        dfr,
        "IN",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u20a8",
        html="&#x20a8;",
    )
    insert(
        "MYR", "458", "MR", 345, "Malaysian Ringgit", dfr, "MY", "ACT/365", "ACT/365"
    )
    insert(
        "VND",
        "704",
        "VD",
        1700000,
        "Vietnamese Dong",
        0,
        "VN",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u20ab",
        html="&#x20ab;",
    )

    # LATIN AMERICA
    insert(
        "BRL",
        "986",
        "BC",
        200,
        "Brazilian Real",
        dfr,
        "BR",
        "BUS/252",
        "BUS/252",
        symbol_raw=r"R$",
    )
    insert(
        "PEN",
        "604",
        "PS",
        220,
        "Peruvian New Sol",
        dfr,
        "PE",
        "ACT/360",
        "ACT/360",
        symbol_raw=r"S/.",
    )
    insert(
        "ARS",
        "032",
        "AP",
        301,
        "Argentine Peso",
        dfr,
        "AR",
        "30/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "MXN",
        "484",
        "MP",
        1330,
        "Mexican Peso",
        dfr,
        "MX",
        "ACT/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "CLP",
        "152",
        "CH",
        54500,
        "Chilean Peso",
        2,
        "CL",
        "ACT/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    insert(
        "COP",
        "170",
        "CL",
        190000,
        "Colombian Peso",
        2,
        "CO",
        "ACT/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    # TODO: Check towletters code and position
    insert(
        "JMD",
        "388",
        "JD",
        410,
        "Jamaican Dollar",
        dfr,
        "JM",
        "ACT/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    # TODO: Check towletters code and position
    insert(
        "TTD",
        "780",
        "TT",
        410,
        "Trinidad and Tobago Dollar",
        dfr,
        "TT",
        "ACT/360",
        "ACT/360",
        symbol_raw=dollar,
        html="&#x24;",
    )
    # TODO: Check towletters code and position
    insert(
        "BMD",
        "060",
        "BD",
        410,
        "Bermudian Dollar",
        dfr,
        "BM",
        symbol_raw=dollar,
        html="&#x24;",
    )

    # EASTERN EUROPE
    insert(
        "CZK",
        "203",
        "CK",
        28,
        "Czech Koruna",
        dfr,
        "CZ",
        "ACT/360",
        "ACT/360",
        symbol_raw=r"\u004b\u010d",
    )
    insert(
        "PLN",
        "985",
        "PZ",
        29,
        "Polish Złoty",
        dfr,
        "PL",
        "ACT/ACT",
        "ACT/365",
        symbol_raw=r"\u007a\u0142",
    )
    insert(
        "TRY",
        "949",
        "TY",
        30,
        "Turkish Lira",
        dfr,
        "TR",
        "ACT/360",
        "ACT/360",
        symbol_raw=r"\u0054\u004c",
    )
    insert(
        "HUF",
        "348",
        "HF",
        32,
        "Hungarian Forint",
        dfr,
        "HU",
        "ACT/365",
        "ACT/360",
        symbol_raw=r"Ft",
        html="Ft",
    )
    insert(
        "RON",
        "946",
        "RN",
        34,
        "Romanian Leu",
        dfr,
        "RO",
        "ACT/360",
        "ACT/360",
    )
    insert(
        "UAH",
        "980",
        "UH",
        35,
        "Ukrainian Hryvnia",
        dfr,
        "UA",
        "ACT/ACT",
        "ACT/ACT",
        symbol_raw=r"\u20b4",
        html="&#x20b4;",
    )
    insert(
        "RUB",
        "643",
        "RR",
        36,
        "Russian Ruble",
        dfr,
        "RU",
        "ACT/ACT",
        "ACT/ACT",
        symbol_raw=r"\u0440\u0443\u0431",
    )
    # TODO: Check towletters code and position
    insert(
        "KZT",
        "398",
        "KT",
        410,
        "Tenge",
        dfr,
        "KZ",
        symbol_raw=r"\u20b8",
        html="&#x20b8;",
    )
    # TODO: Check towletters code and position
    insert(
        "BGN",
        "975",
        "BN",
        410,
        "Bulgarian Lev",
        dfr,
        "BG",
        symbol_raw=r"\u043b\u0432.",
        html="&#1083;&#1074;",
    )

    # MIDDLE EAST & AFRICA
    insert(
        "ILS",
        "376",
        "IS",
        410,
        "Israeli Shekel",
        dfr,
        "IL",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"\u20aa",
        html="&#x20aa;",
    )
    # TODO: Check towletters code and position
    insert(
        "AED",
        "784",
        "AE",
        410,
        "United Arab Emirates Dirham",
        dfr,
        "AE",
    )
    # TODO: Check towletters code and position
    insert(
        "QAR",
        "634",
        "QA",
        410,
        "Qatari Riyal",
        dfr,
        "QA",
        symbol_raw=r"\ufdfc",
        html="&#xfdfc;",
    )
    # TODO: Check towletters code and position
    insert(
        "SAR",
        "682",
        "SR",
        410,
        "Saudi Riyal",
        dfr,
        "SA",
        symbol_raw=r"\ufdfc",
        html="&#xfdfc;",
    )
    insert(
        "EGP",
        "818",
        "EP",
        550,
        "Egyptian Pound",
        dfr,
        "EG",
        symbol_raw=r"\u00a3",
        html="&#xa3;",
    )
    insert(
        "NGN",
        "566",
        "NG",
        650,
        "Nigerian Naira",
        dfr,
        "NG",
        symbol_raw=r"\u20a6",
        html="&#x20A6;",
    )
    insert(
        "ZAR",
        "710",
        "SA",
        750,
        "South African Rand",
        dfr,
        "ZA",
        "ACT/365",
        "ACT/365",
        symbol_raw=r"R",
        html="R",
    )

    # BITCOIN
    insert(
        "XBT",
        "000",
        "BT",
        -1,
        "Bitcoin",
        8,
        "WW",
        symbol_raw=r"\u0e3f",
        html="&#xe3f;",
    )
